package ru.dbotthepony.kstarbound.gl

import com.google.common.collect.ImmutableList
import org.lwjgl.opengl.GL46.*

enum class GLType(val identity: Int, val typeIndentity: Int, val byteSize: Int, val logicalSize: Int) {
	INT(GL_INT, GL_INT, 4, 1),
	UINT(GL_UNSIGNED_INT, GL_UNSIGNED_INT, 4, 1),
	FLOAT(GL_FLOAT, GL_FLOAT, 4, 1),
	DOUBLE(GL_DOUBLE, GL_DOUBLE, 8, 1),

	VEC2F(GL_FLOAT_VEC2, GL_FLOAT, 8, 2),
	VEC3F(GL_FLOAT_VEC3, GL_FLOAT, 12, 3),
	VEC4F(GL_FLOAT_VEC4, GL_FLOAT, 16, 4),

	VEC2I(GL_INT_VEC2, GL_INT, 8, 2),
	VEC3I(GL_INT_VEC3, GL_INT, 12, 3),
	VEC4I(GL_INT_VEC4, GL_INT, 16, 4),

	MAT2F(GL_FLOAT_MAT2, GL_FLOAT, 2 * 2 * 4, 2 * 2),
	MAT3F(GL_FLOAT_MAT3, GL_FLOAT, 3 * 3 * 4, 3 * 3),
	MAT4F(GL_FLOAT_MAT4, GL_FLOAT, 4 * 4 * 4, 4 * 4),
}

interface IGLAttributeList {
	fun apply(target: GLVertexArrayObject, enable: Boolean = false)
}

data class AttributeListPosition(val name: String, val index: Int, val glType: GLType, val stride: Int, val offset: Long)

/**
 * Хранит список аттрибутов для применения к Vertex Array Object
 *
 * Аттрибуты плотно упакованы и идут один за другим
 *
 * Создаётся через [GLFlatAttributeListBuilder]
 */
class GLFlatAttributeList(builder: GLFlatAttributeListBuilder) : IGLAttributeList {
	val attributes: List<AttributeListPosition>
	val size get() = attributes.size
	val stride: Int

	operator fun get(index: Int) = attributes[index]

	fun vertexBuilder(vertexType: VertexType) = VertexBuilder(this, vertexType)

	init {
		val buildList = ArrayList<AttributeListPosition>()

		var offset = 0L
		var stride = 0

		for (i in builder.attributes) {
			stride += i.second.byteSize
		}

		this.stride = stride

		// val counter = mutableMapOf<Int, Int>()

		for (i in builder.attributes.indices) {
			val value = builder.attributes[i].second
			buildList.add(AttributeListPosition(builder.attributes[i].first, i, value, stride, offset))
			offset += value.byteSize

			// counter[value.typeIndentity] = (counter[value.typeIndentity] ?: 0) + 1
		}

		attributes = ImmutableList.copyOf(buildList)
	}

	override fun apply(target: GLVertexArrayObject, enable: Boolean) {
		for (i in attributes.indices) {
			val value = attributes[i]
			target.attribute(i, value.glType.logicalSize, value.glType.typeIndentity, false, value.stride, value.offset)

			if (enable) {
				target.enableAttribute(i)
			}
		}
	}

	companion object {
		val VEC3F = GLFlatAttributeListBuilder().also {it.push(GLType.VEC3F)}.build()
		val VERTEX_TEXTURE = GLFlatAttributeListBuilder().also {it.push(GLType.VEC3F).push(GLType.VEC2F)}.build()
		val VERTEX_2D_TEXTURE = GLFlatAttributeListBuilder().also {it.push(GLType.VEC2F).push(GLType.VEC2F)}.build()
	}
}

class GLFlatAttributeListBuilder : IGLAttributeList {
	val attributes = ArrayList<Pair<String, GLType>>()

	private fun findName(name: String): Boolean {
		for (value in attributes) {
			if (value.first == name) {
				return true
			}
		}

		return false
	}

	fun push(type: GLType): GLFlatAttributeListBuilder {
		return push("$type#${attributes.size}", type)
	}

	fun push(name: String, type: GLType): GLFlatAttributeListBuilder {
		check(!findName(name)) { "Already has named attribute $name!" }
		attributes.add(name to type)
		return this
	}

	fun build() = GLFlatAttributeList(this)

	@Deprecated("Используй build()")
	override fun apply(target: GLVertexArrayObject, enable: Boolean) {
		var offset = 0L
		var stride = 0

		for (i in attributes) {
			stride += i.second.byteSize
		}

		for (i in attributes.indices) {
			val value = attributes[i].second
			target.attribute(i, value.logicalSize, value.typeIndentity, false, stride, offset)
			offset += value.byteSize

			if (enable) {
				target.enableAttribute(i)
			}
		}
	}

	companion object {
		val VEC3F = GLFlatAttributeList.VEC3F
		val VERTEX_TEXTURE = GLFlatAttributeList.VERTEX_TEXTURE
	}
}
