package ru.dbotthepony.kstarbound.gl

import org.lwjgl.opengl.GL46.*
import java.nio.ByteBuffer
import java.nio.ByteOrder

enum class VertexType(val elements: Int, val indicies: IntArray) {
	TRIANGLES(3, intArrayOf(0, 1, 2)),
	QUADS(4, intArrayOf(0, 1, 2, 1, 2, 3))
}

typealias VertexTransformer = (VertexBuilder.Vertex, Int) -> VertexBuilder.Vertex
private val emptyTransform: VertexTransformer = { it, _ -> it }

object VertexTransformers {
	fun uv(u0: Float,
		   v0: Float,
		   u1: Float,
		   v1: Float,
		   lambda: VertexTransformer = emptyTransform): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(u0, v0)
				1 -> it.pushVec2f(u1, v0)
				2 -> it.pushVec2f(u0, v1)
				3 -> it.pushVec2f(u1, v1)
			}

			return@transformer lambda(it, index)
		}
	}

	fun uv(): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(0f, 0f)
				1 -> it.pushVec2f(1f, 0f)
				2 -> it.pushVec2f(0f, 1f)
				3 -> it.pushVec2f(1f, 1f)
			}

			return@transformer it
		}
	}

	fun uv(lambda: VertexTransformer): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(0f, 0f)
				1 -> it.pushVec2f(1f, 0f)
				2 -> it.pushVec2f(0f, 1f)
				3 -> it.pushVec2f(1f, 1f)
			}

			return@transformer lambda(it, index)
		}
	}
}

class VertexBuilder(val attributes: GLFlatAttributeList, private val type: VertexType) {
	private val verticies = ArrayList<Vertex>()
	val indexCount get() = (verticies.size / type.elements) * type.indicies.size

	fun begin(): VertexBuilder {
		verticies.clear()
		return this
	}

	fun vertex(): Vertex {
		return Vertex()
	}

	fun quadZ(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		z: Float,
		lambda: VertexTransformer = emptyTransform
	): VertexBuilder {
		check(type == VertexType.QUADS) { "Currently building $type" }

		lambda(Vertex().pushVec3f(x0, y0, z), 0).end()
		lambda(Vertex().pushVec3f(x1, y0, z), 1).end()
		lambda(Vertex().pushVec3f(x0, y1, z), 2).end()
		lambda(Vertex().pushVec3f(x1, y1, z), 3).end()

		return this
	}

	fun quad(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		lambda: VertexTransformer = emptyTransform
	): VertexBuilder {
		check(type == VertexType.QUADS) { "Currently building $type" }

		lambda(Vertex().pushVec2f(x0, y0), 0).end()
		lambda(Vertex().pushVec2f(x1, y0), 1).end()
		lambda(Vertex().pushVec2f(x0, y1), 2).end()
		lambda(Vertex().pushVec2f(x1, y1), 3).end()

		return this
	}

	fun checkValid() {
		for (vertex in verticies) {
			vertex.checkValid()
		}
	}

	/**
	 * Загружает буфер в указанные VBO и EBO
	 *
	 * операция создаёт мусор вне кучи и довольно медленная
	 */
	fun upload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int = GL_DYNAMIC_DRAW) {
		require(vbo.isArray) { "$vbo is not an array" }
		require(ebo.isElementArray) { "$vbo is not an element array" }

		checkValid()

		check(verticies.size % type.elements == 0) { "Not fully built (expected ${type.elements} verticies to be present for each element, last element has only ${verticies.size % type.elements})" }

		vbo.bind()
		ebo.bind()

		if (verticies.size == 0) {
			vbo.bufferData(intArrayOf(), drawType)
			ebo.bufferData(intArrayOf(), drawType)
			return
		}

		val bytes = ByteBuffer.allocateDirect(verticies.size * attributes.stride)
		bytes.order(ByteOrder.nativeOrder())

		for (vertex in verticies) {
			vertex.upload(bytes)
		}

		check(bytes.position() == bytes.capacity()) { "Buffer is not fully filled (position: ${bytes.position()}; capacity: ${bytes.capacity()})" }

		bytes.position(0)
		vbo.bufferData(bytes, drawType)

		val elementIndicies = IntArray((verticies.size / type.elements) * type.indicies.size)
		var offset = 0
		var offsetVertex = 0

		for (i in 0 until verticies.size / type.elements) {
			for (i2 in type.indicies.indices) {
				elementIndicies[offset + i2] = type.indicies[i2] + offsetVertex
			}

			offset += type.indicies.size
			offsetVertex += type.elements
		}

		ebo.bufferData(elementIndicies, drawType)
	}

	inner class Vertex {
		init {
			verticies.add(this)
		}

		private val store = arrayOfNulls<Any>(attributes.size)
		private var index = 0

		fun upload(bytes: ByteBuffer) {
			for (element in store) {
				when (element) {
					is FloatArray -> for (i in element) bytes.putFloat(i)
					is IntArray -> for (i in element) bytes.putInt(i)
					is ByteArray -> for (i in element) bytes.put(i)
					is DoubleArray -> for (i in element) bytes.putDouble(i)
					else -> throw IllegalStateException("Unknown element $element")
				}
			}
		}

		override fun toString(): String {
			return "Vertex(${store.map {
				return@map when (it) {
					is FloatArray -> it.joinToString(", ")
					is IntArray -> it.joinToString(", ")
					is ByteArray -> it.joinToString(", ")
					is DoubleArray -> it.joinToString(", ")
					else -> "null"
				} }.joinToString("; ")})"
		}

		fun expect(name: String): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected $name")
			}

			if (attributes[index].name != name) {
				throw IllegalStateException("Expected $name, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		fun expect(type: GLType): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected type $type")
			}

			if (attributes[index].glType != type) {
				throw IllegalStateException("Expected $type, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		fun pushVec3f(x: Float, y: Float, z: Float): Vertex {
			expect(GLType.VEC3F)
			store[index++] = floatArrayOf(x, y, z)
			return this
		}

		fun pushVec2f(x: Float, y: Float): Vertex {
			expect(GLType.VEC2F)
			store[index++] = floatArrayOf(x, y)
			return this
		}

		fun checkValid() {
			for (elem in store.indices) {
				if (store[elem] == null) {
					throw IllegalStateException("Vertex element at position $elem is null")
				}
			}
		}

		fun end(): VertexBuilder {
			checkValid()
			return this@VertexBuilder
		}
	}
}

