package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.math.IVector2i
import ru.dbotthepony.kstarbound.math.Vector2i

/**
 * Представляет из себя класс, который содержит состояние тайла на заданной позиции
 */
data class ChunkTile(val def: TileDefinition) {
	var color = -1
	var forceVariant = -1
}

interface ITileMap {
	/**
	 * Относительная проверка находится ли координата вне границ чанка
	 */
	fun isOutside(x: Int, y: Int): Boolean {
		return x !in 0 until CHUNK_SIZE || y !in 0 until CHUNK_SIZE
	}
}

/**
 * Предоставляет интерфейс для доступа к тайлам в чанке
 */
interface ITileGetter : ITileMap {
	/**
	 * Возвращает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun get(x: Int, y: Int): ChunkTile?

	/**
	 * Возвращает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun get(pos: Vector2i) = get(pos.x, pos.y)

	/**
	 * Возвращает итератор пар <Vector2i, Тайл?>
	 *
	 * Вектор имеет ОТНОСИТЕЛЬНЫЕ значения внутри самого чанка
	 */
	val posToTile: Iterator<Pair<Vector2i, ChunkTile?>> get() {
		return object : Iterator<Pair<Vector2i, ChunkTile?>> {
			private var x = 0
			private var y = 0

			private fun idx() = x + CHUNK_SIZE * y

			override fun hasNext(): Boolean {
				return idx() < CHUNK_SIZE * CHUNK_SIZE
			}

			override fun next(): Pair<Vector2i, ChunkTile?> {
				if (!hasNext()) {
					throw IllegalStateException("Already iterated everything!")
				}

				val tile = this@ITileGetter[x, y]
				val pos = Vector2i(x, y)

				x++

				if (x >= CHUNK_SIZE) {
					y++
					x = 0
				}

				return pos to tile
			}
		}
	}
}

/**
 * Интерфейс предоставляет из себя описание класса, который имеет координаты чанка
 */
interface IChunkPositionable : ITileMap {
	val pos: ChunkPos

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(x: Int, y: Int): Long {
		var long = (x or (pos.x shl CHUNK_SHIFT)) * 738548L + (y or (pos.y shl CHUNK_SHIFT)) * 2191293543L
		long = long xor 8339437585692L
		long = (long ushr 4) or (long shl 52)
		long *= 7848344324L
		long = (long ushr 12) or (long shl 44)
		return long
	}

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(x: Int, y: Int): Double {
		return (randomLongFor(x, y) / 9.223372036854776E18) / 2.0 + 0.5
	}

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(pos: Vector2i) = randomLongFor(pos.x, pos.y)

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(pos: Vector2i) = randomDoubleFor(pos.x, pos.y)
}

/**
 * Предоставляет интерфейс по установке тайлов в чанке
 */
interface ITileSetter : ITileMap {
	/**
	 * Устанавливает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile?
	/**
	 * Устанавливает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun set(pos: Vector2i, tile: TileDefinition?) = set(pos.x, pos.y, tile)
}

interface ITileGetterSetter : ITileGetter, ITileSetter
interface ITileChunk : ITileGetter, IChunkPositionable
interface IMutableTileChunk : ITileChunk, ITileSetter

const val CHUNK_SHIFT = 6
const val CHUNK_SIZE = 1 shl CHUNK_SHIFT // 64
const val CHUNK_SIZE_FF = CHUNK_SIZE - 1

data class ChunkPos(override val x: Int, override val y: Int) : IVector2i<ChunkPos>() {
	constructor(pos: Vector2i) : this(pos.x shr CHUNK_SHIFT, pos.y shr CHUNK_SHIFT)
	override fun make(x: Int, y: Int) = ChunkPos(x, y)

	val firstBlock get() = Vector2i(x shl CHUNK_SHIFT, y shl CHUNK_SHIFT)
	val lastBlock get() = Vector2i(((x + 1) shl CHUNK_SHIFT) - 1, ((y + 1) shl CHUNK_SHIFT) - 1)
}

/**
 * Предоставляет доступ к чанку и его соседям
 *
 * В основном для использования в местах, где нужен не мир, а определённый чанк мира,
 * и при этом координаты проверяются относительно чанка и могут спокойно выйти за его пределы,
 * с желанием получить тайл из соседнего чанка
 */
open class TileChunkView(
	open val center: ITileChunk,

	open val right: ITileChunk?,
	open val top: ITileChunk?,
	open val topRight: ITileChunk?,
	open val topLeft: ITileChunk?,

	open val left: ITileChunk?,
	open val bottom: ITileChunk?,
	open val bottomLeft: ITileChunk?,
	open val bottomRight: ITileChunk?,
) : ITileChunk {
	override fun get(x: Int, y: Int): ChunkTile? {
		if (x in 0 .. CHUNK_SIZE_FF) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return center[x, y]
			}

			if (y < 0) {
				return bottom?.get(x, y + CHUNK_SIZE)
			} else {
				return top?.get(x, y - CHUNK_SIZE)
			}
		}

		if (x < 0) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return left?.get(x + CHUNK_SIZE, y)
			}

			if (y < 0) {
				return bottomLeft?.get(x + CHUNK_SIZE, y + CHUNK_SIZE)
			} else {
				return topLeft?.get(x + CHUNK_SIZE, y - CHUNK_SIZE)
			}
		} else {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return right?.get(x - CHUNK_SIZE, y)
			}

			if (y < 0) {
				return bottomRight?.get(x - CHUNK_SIZE, y + CHUNK_SIZE)
			} else {
				return topRight?.get(x - CHUNK_SIZE, y - CHUNK_SIZE)
			}
		}
	}

	override val pos: ChunkPos
		get() = center.pos
}

class MutableTileChunkView(
	override val center: IMutableTileChunk,

	override val right: IMutableTileChunk?,
	override val top: IMutableTileChunk?,
	override val topRight: IMutableTileChunk?,
	override val topLeft: IMutableTileChunk?,

	override val left: IMutableTileChunk?,
	override val bottom: IMutableTileChunk?,
	override val bottomLeft: IMutableTileChunk?,
	override val bottomRight: IMutableTileChunk?,
) : TileChunkView(center, right, top, topRight, topLeft, left, bottom, bottomLeft, bottomRight), IMutableTileChunk {
	override fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? {
		if (x in 0 .. CHUNK_SIZE_FF) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return center.set(x, y, tile)
			}

			if (y < 0) {
				return bottom?.set(x, y + CHUNK_SIZE, tile)
			} else {
				return top?.set(x, y - CHUNK_SIZE, tile)
			}
		}

		if (x < 0) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return left?.set(x + CHUNK_SIZE, y, tile)
			}

			if (y < 0) {
				return bottomLeft?.set(x + CHUNK_SIZE, y + CHUNK_SIZE, tile)
			} else {
				return topLeft?.set(x + CHUNK_SIZE, y - CHUNK_SIZE, tile)
			}
		} else {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return right?.set(x - CHUNK_SIZE, y, tile)
			}

			if (y < 0) {
				return bottomRight?.set(x - CHUNK_SIZE, y + CHUNK_SIZE, tile)
			} else {
				return topRight?.set(x - CHUNK_SIZE, y - CHUNK_SIZE, tile)
			}
		}
	}
}

open class Chunk(val world: World, val pos: ChunkPos) {
	inner class ChunkTileLayer : IMutableTileChunk {
		override val pos: ChunkPos
			get() = this@Chunk.pos

		/**
		 * Хранит тайлы как x + y * CHUNK_SIZE
		 */
		private val tiles = arrayOfNulls<ChunkTile>(CHUNK_SIZE * CHUNK_SIZE)

		override operator fun get(x: Int, y: Int): ChunkTile? {
			if (isOutside(x, y))
				return null

			return tiles[x or (y shl CHUNK_SHIFT)]
		}

		operator fun set(x: Int, y: Int, tile: ChunkTile?) {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.def?.materialName} at $x $y, but that is outside of chunk's range")

			tiles[x or (y shl CHUNK_SHIFT)] = tile
		}

		override operator fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.materialName} at $x $y, but that is outside of chunk's range")

			val chunkTile = if (tile != null) ChunkTile(tile) else null
			this[x, y] = chunkTile
			return chunkTile
		}

		override fun randomLongFor(x: Int, y: Int): Long {
			return super.randomLongFor(x, y) xor world.seed
		}
	}

	val foreground = ChunkTileLayer()
	val background = ChunkTileLayer()

	companion object {
		val EMPTY = object : IMutableTileChunk {
			override val pos = ChunkPos(0, 0)

			override fun get(x: Int, y: Int): ChunkTile? = null
			override fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? = null
		}
	}
}
